#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

all = ["local_hardware_info", "VersionTable"]

import distutils.core
import os
import platform
import sys
import time

import pkg_resources
import psutil
from IPython.core.magic import Magics, line_magic, magics_class
from IPython.display import HTML, display

import nannos

dir_path = os.path.dirname(os.path.realpath(__file__))


def local_hardware_info():
    """Basic hardware information about the local machine.
    Gives actual number of CPU's in the machine, even when hyperthreading is
    turned on. CPU count defaults to 1 when true count can't be determined.
    Returns:
        dict: The hardware information.
    """
    results = {
        "python_compiler": platform.python_compiler(),
        "python_build": ", ".join(platform.python_build()),
        "python_version": platform.python_version(),
        "os": platform.system(),
        "memory": psutil.virtual_memory().total / (1024**3),
        "cpus": psutil.cpu_count(logical=False) or 1,
    }
    return results


@magics_class
class VersionTable(Magics):
    """A class of status magic functions."""

    @line_magic
    def nannos_version_table(self, line="", cell=None):
        """
        Print an HTML-formatted table with version numbers for Nannos and its
        dependencies. This should make it possible to reproduce the environment
        and the calculation later on.
        """

        html = "<h3>Version Information</h3>"
        html += "<table>"
        html += "<tr><th>Package</th></tr>"

        packages = []

        packages_names = ["nannos", "numpy", "scipy", "matplotlib"]

        for pkg in packages_names:
            ver = pkg_resources.get_distribution(pkg).version

            packages.append((f"<code>{pkg}</code>", ver))

        ver = pkg_resources.get_distribution("autograd_latest").version
        packages.append(("<code>autograd</code>", ver))

        for name, version in packages:
            html += f"<tr><td>{name}</td><td>{version}</td></tr>"

        html += "<tr><th>System information</th></tr>"

        local_hw_info = local_hardware_info()
        sys_info = [
            ("Python version", local_hw_info["python_version"]),
            ("Python compiler", local_hw_info["python_compiler"]),
            ("Python build", local_hw_info["python_build"]),
            ("OS", "%s" % local_hw_info["os"]),
            ("CPUs", "%s" % local_hw_info["cpus"]),
            ("Memory (Gb)", "%s" % local_hw_info["memory"]),
        ]

        for name, version in sys_info:
            html += f"<tr><td>{name}</td><td>{version}</td></tr>"

        html += "<tr><td colspan='2'>%s</td></tr>" % time.strftime(
            "%a %b %d %H:%M:%S %Y %Z"
        )
        html += "</table>"

        return display(HTML(html))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


from .. import backend as bk
from .. import get_backend
from ..constants import pi
from ..utils import norm

_arctan2 = bk.atan2 if get_backend() == "torch" else bk.arctan2


def get_jones_field(t):
    norm_t = norm(t)
    n = [-t[1], t[0]]
    theta = _arctan2(t[1], t[0])
    phi = pi / 8 * (1 + bk.cos(pi * norm_t))
    expo = bk.exp(1j * theta)
    J = [
        expo / norm_t * (t[i] * bk.cos(phi) + 1j * n[i] * bk.sin(phi)) for i in range(2)
    ]
    return J



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

import os

from .__about__ import __author__, __description__, __version__
from .__about__ import data as _data
from .log import *

available_backends = ["numpy", "scipy", "autograd", "jax", "torch"]


def print_info():
    print(f"nannos v{__version__}")
    print("=============")
    print(__description__)
    print(f"Author: {__author__}")
    print(f"Licence: {_data['License']}")


def has_torch():
    try:
        import torch

        return True
    except ModuleNotFoundError:
        return False


def has_jax():
    try:
        import jax

        return True
    except ModuleNotFoundError:
        return False


def _has_cuda():
    try:
        import torch

        return torch.cuda.is_available()
    except ModuleNotFoundError:
        return False


HAS_TORCH = has_torch()
HAS_CUDA = _has_cuda()

HAS_JAX = has_jax()


def use_gpu(boolean):
    global DEVICE
    global _CPU_DEVICE
    global _GPU_DEVICE

    if boolean:

        if BACKEND not in ["torch"]:
            logger.debug(f"Cannot use GPU with {BACKEND} backend.")
            _delvar("_GPU_DEVICE")
            _CPU_DEVICE = True
        else:
            if not HAS_TORCH:
                logger.warning("pytorch not found. Cannot use GPU.")
                _delvar("_GPU_DEVICE")
                _CPU_DEVICE = True
            elif not HAS_CUDA:
                logger.warning("cuda not found. Cannot use GPU.")
                _delvar("_GPU_DEVICE")
                _CPU_DEVICE = True
            else:
                DEVICE = "cuda"
                logger.debug("Using GPU.")
                _delvar("_CPU_DEVICE")
                _GPU_DEVICE = True
    else:
        _CPU_DEVICE = True
        _delvar("_GPU_DEVICE")
        DEVICE = "cpu"
        logger.debug("Using CPU.")
    _reload_package()


def jit(fun, **kwargs):
    if BACKEND == "jax":
        from jax import jit

        return jit(fun, **kwargs)
    else:
        return fun


def _delvar(VAR):
    if VAR in globals():
        del globals()[VAR]


def set_backend(backend):
    """Set the numerical backend.

    Parameters
    ----------
    backend : str
        Either ``numpy``, ``scipy``, ``autograd``, ``torch`` or ``jax``.


    """

    global _NUMPY
    global _SCIPY
    global _AUTOGRAD
    global _JAX
    global _TORCH
    global _FORCE_BACKEND

    _FORCE_BACKEND = 1

    if backend == get_backend():
        return
    #
    # _backend_env_var = os.environ.get("NANNOS_BACKEND")
    # if _backend_env_var is not None:
    #     if _backend_env_var in available_backends:
    #         if backend != _backend_env_var:
    #             # _delvar("_FORCE_BACKEND")
    #             pass
    #         else:
    #             backend = _backend_env_var

    if backend == "autograd":
        logger.debug("Setting autograd backend")
        _AUTOGRAD = True
        _delvar("_JAX")
        _delvar("_TORCH")
        _delvar("_SCIPY")
    elif backend == "scipy":
        logger.debug("Setting scipy backend")
        _SCIPY = True
        _delvar("_JAX")
        _delvar("_TORCH")
        _delvar("_AUTOGRAD")
    elif backend == "jax":
        logger.debug("Setting jax backend")
        _JAX = True
        _delvar("_SCIPY")
        _delvar("_TORCH")
        _delvar("_AUTOGRAD")
    elif backend == "torch":
        _TORCH = True
        logger.debug("Setting torch backend")
        _delvar("_SCIPY")
        _delvar("_JAX")
        _delvar("_AUTOGRAD")
    elif backend == "numpy":
        _NUMPY = True
        logger.debug("Setting numpy backend")
        _delvar("_SCIPY")
        _delvar("_JAX")
        _delvar("_AUTOGRAD")
        _delvar("_TORCH")
    else:
        raise ValueError(
            f"Unknown backend '{backend}'. Please choose between 'numpy', 'scipy', 'jax', 'torch' and 'autograd'."
        )
    _reload_package()


def _reload_package():

    import importlib
    import sys

    import nannos

    importlib.reload(nannos)

    its = [s for s in sys.modules.items() if s[0].startswith("nannos")]
    for k, v in its:
        importlib.reload(v)


def get_backend():

    if "_SCIPY" in globals():
        return "scipy"
    elif "_AUTOGRAD" in globals():
        return "autograd"
    elif "_JAX" in globals():
        return "jax"
    elif "_TORCH" in globals():
        return "torch"
    else:
        return "numpy"


def _grad(f):
    raise NotImplementedError(f"grad is not implemented for {BACKEND} backend.")


if "_SCIPY" in globals():

    import numpy

    grad = _grad
    backend = numpy
elif "_AUTOGRAD" in globals():
    from autograd import grad, numpy

    backend = numpy
elif "_JAX" in globals():
    if HAS_JAX:

        from jax.config import config

        config.update("jax_platform_name", "cpu")
        config.update("jax_enable_x64", True)

        # TODO: jax eig not implemented on GPU
        # see https://github.com/google/jax/issues/1259

        # TODO: support jax properly (is it faster than autograd? use jit?)
        # jax does not support eig
        # for autodif wrt eigenvectors yet.
        # see: https://github.com/google/jax/issues/2748

        # if DEVICE == "cpu":
        #     config.update("jax_platform_name", "cpu")
        # else:
        #     config.update("jax_platform_name", "gpu")
        # from jax import grad, numpy
        from jax import numpy

        grad = _grad
        backend = numpy

    else:
        logger.warning("jax not found. Falling back to default numpy backend.")
        set_backend("numpy")
elif "_TORCH" in globals():
    if HAS_TORCH:
        import numpy
        import torch

        # torch.set_default_tensor_type(torch.cuda.FloatTensor)

        backend = torch

        def _array(a, **kwargs):
            if isinstance(a, backend.Tensor):
                return a.to(torch.device(DEVICE))
            else:
                return backend.tensor(a, **kwargs).to(torch.device(DEVICE))

        backend.array = _array

        def grad(f):
            def df(x, *args, **kwargs):
                x = backend.array(x, dtype=bk.float64)
                _x = x.clone().detach().requires_grad_(True)
                out = backend.autograd.grad(
                    f(_x, *args, **kwargs), _x, allow_unused=True
                )[0]
                return out

            return df

    else:
        logger.warning("pytorch not found. Falling back to default numpy backend.")
        set_backend("numpy")
else:
    import numpy

    grad = _grad
    backend = numpy


def get_device():
    return "cuda" if "_GPU_DEVICE" in globals() else "cpu"


BACKEND = get_backend()
DEVICE = get_device()

_backend_env_var = os.environ.get("NANNOS_BACKEND")

if _backend_env_var in available_backends and _backend_env_var is not None:
    if BACKEND != _backend_env_var and "_FORCE_BACKEND" not in globals():
        logger.debug(f"Found environment variable NANNOS_BACKEND={_backend_env_var}")
        set_backend(_backend_env_var)

from . import optimize
from .constants import *
from .excitation import *
from .lattice import *
from .parallel import *
from .sample import *
from .simulation import *
from .utils import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


"""
Geometry helpers
"""

# __all__ = ["shape_mask"]

import shapely.affinity as sa
import shapely.geometry as sg

from . import backend as bk
from . import get_backend

# from https://gist.github.com/perrette/a78f99b76aed54b6babf3597e0b331f8


def _grid_bbox(x, y):
    dx = dy = 0
    return x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2


def _bbox_to_rect(bbox):
    left, right, bottom, top = bbox
    return sg.Polygon([(left, bottom), (right, bottom), (right, top), (left, top)])


def set_index_2d(mat, val, idx1=(None, None), idx2=(None, None)):
    idx = slice(*idx1), slice(*idx2)
    if get_backend() == "jax":
        mat = mat.at[idx].set(val)
    else:
        mat[idx] = val
    return mat


def shape_mask(shp, x, y, m=None):
    """Use recursive sub-division of space and shapely contains method to create a raster mask on a regular grid.

    Parameters
    ----------
    shp : shapely's Polygon (or whatever with a "contains" method and intersects method)
    x, y : 1-D numpy arrays defining a regular grid
    m : mask to fill, optional (will be created otherwise)

    Returns
    -------
    m : boolean 2-D array, True inside shape.

    Examples
    --------
    >>> from shapely.geometry import Point
    >>> poly = Point(0,0).buffer(1)
    >>> x = np.linspace(-5,5,100)
    >>> y = np.linspace(-5,5,100)
    >>> mask = shape_mask(poly, x, y)
    """
    rect = _bbox_to_rect(_grid_bbox(x, y))

    if m is None:
        m = bk.zeros((len(y), len(x)), dtype=bool)

    if not shp.intersects(rect):
        m = set_index_2d(m, False)

    elif shp.contains(rect):
        m = set_index_2d(m, True)

    else:
        lx, ly = m.shape

        if lx == 1 and ly == 1:

            val = shp.contains(sg.Point(x[0], y[0]))
            m = set_index_2d(m, val)

        elif lx == 1:

            val = shape_mask(shp, x[: ly // 2], y, m[:, : ly // 2])
            m = set_index_2d(m, val, idx2=(None, ly // 2))
            val = shape_mask(shp, x[ly // 2 :], y, m[:, ly // 2 :])
            m = set_index_2d(m, val, idx2=(ly // 2, None))

        elif ly == 1:
            val = shape_mask(shp, x, y[: lx // 2], m[: lx // 2])
            m = set_index_2d(m, val, idx1=(None, lx // 2))
            val = shape_mask(shp, x, y[lx // 2 :], m[lx // 2 :])
            m = set_index_2d(m, val, idx1=(lx // 2, None))

        else:
            val = shape_mask(shp, x[: ly // 2], y[: lx // 2], m[: lx // 2, : ly // 2])

            m = set_index_2d(m, val, (None, lx // 2), (None, ly // 2))
            val = shape_mask(shp, x[ly // 2 :], y[: lx // 2], m[: lx // 2, ly // 2 :])
            m = set_index_2d(m, val, (None, lx // 2), (ly // 2, None))
            val = shape_mask(shp, x[: ly // 2], y[lx // 2 :], m[lx // 2 :, : ly // 2])
            m = set_index_2d(m, val, (lx // 2, None), (None, ly // 2))
            val = shape_mask(shp, x[ly // 2 :], y[lx // 2 :], m[lx // 2 :, ly // 2 :])
            m = set_index_2d(m, val, (lx // 2, None), (ly // 2, None))

    return m


def geometry_mask(geom, lattice, Nx, Ny):
    x0 = bk.linspace(0, 1.0, Nx)
    y0 = bk.linspace(0, 1.0, Ny)
    x_, y_ = bk.meshgrid(x0, y0, indexing="ij")
    grid = bk.stack([x_, y_])
    x, y = grid[0][:, 0], grid[1][0, :]
    invM = bk.linalg.inv(lattice.matrix)
    matrix = invM.ravel().tolist() + [0, 0]
    geom = sa.affine_transform(geom, matrix)
    mask = shape_mask(geom, x, y).T
    return mask


def polygon(vertices, lattice, Nx, Ny):
    polygon = sg.Polygon(vertices)
    return geometry_mask(polygon, lattice, Nx, Ny)


def circle(center, radius, lattice, Nx, Ny):
    circ = sg.Point(*center).buffer(radius)
    return geometry_mask(circ, lattice, Nx, Ny)


def ellipse(center, radii, lattice, Nx, Ny, rotate=0):
    radius_x, radius_y = radii
    cent = sg.Point(*center)
    circ = cent.buffer(radius_x)
    ell = sa.scale(circ, xfact=1.0, yfact=radius_y / radius_x, zfact=1.0, origin=cent)
    if rotate != 0:
        ell = sa.rotate(ell, rotate, origin=cent, use_radians=False)
    return geometry_mask(ell, lattice, Nx, Ny)


def rectangle(center, widths, lattice, Nx, Ny, rotate=0):
    vertices = [[center[0] - widths[0] / 2, center[1] - widths[1] / 2]]
    vertices.append([center[0] + widths[0] / 2, center[1] - widths[1] / 2])
    vertices.append([center[0] + widths[0] / 2, center[1] + widths[1] / 2])
    vertices.append([center[0] - widths[0] / 2, center[1] + widths[1] / 2])
    rect = sg.Polygon(vertices)
    cent = sg.Point(*center)
    if rotate != 0:
        rect = sa.rotate(rect, rotate, origin=cent, use_radians=False)
    return geometry_mask(rect, lattice, Nx, Ny)


def square(center, width, lattice, Nx, Ny, rotate=0):
    return rectangle(center, (width, width), lattice, Nx, Ny, rotate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


__all__ = ["Layer"]


from copy import copy

from . import backend as bk
from . import jit
from .formulations.jones import get_jones_field
from .formulations.tangent import get_tangent_field
from .plot import *
from .utils import block
from .utils.helpers import _reseter


class Layer:
    """A layer object.

    Parameters
    ----------
    name : str
        Name of the layer (the default is "layer").
    thickness : float
        Thickness of the layer (the default is 0). Must be positive.
    epsilon : complex
        Permittivity `epsilon` (the default is 1).
    mu : complex
        Permeability `mu` (the default is 1).


    """

    def __init__(
        self,
        name="layer",
        thickness=0,
        epsilon=1,
        mu=1,
        lattice=None,
        tangent_field=None,
        tangent_field_type="fft",
    ):
        if thickness is not None:
            if thickness < 0:
                raise ValueError("thickness must be positive.")
        self.name = name
        self.thickness = thickness
        self.epsilon = bk.array(epsilon, dtype=bk.complex128)
        self.mu = bk.array(mu, dtype=bk.complex128)
        self.lattice = lattice
        self.tangent_field = tangent_field
        self.tangent_field_type = tangent_field_type
        self.iscopy = False
        self.original = self

        if self.mu != 1:
            # TODO: case with magnetic materials
            raise NotImplementedError(
                "Permeability different than unity not yet implemented."
            )

    def __repr__(self):
        return f"{self.name}"

    def plot(
        self,
        nper=1,
        ax=None,
        cmap="tab20c",
        show_cell=False,
        comp="re",
        cellstyle="w-",
        **kwargs,
    ):
        lattice = self.lattice
        if comp not in ["re", "im"]:
            raise ValueError(f"Unknown component {comp}. must be `re` or `im`.")
        toplot = self.epsilon.real if comp == "re" else self.epsilon.imag
        return plot_layer(
            lattice,
            lattice.grid,
            toplot,
            nper,
            ax,
            cmap,
            show_cell,
            cellstyle,
            **kwargs,
        )

    @property
    def is_solved(self):
        return hasattr(self, "eigenvalues") and hasattr(self, "eigenvectors")

    def reset(self, param="all"):
        if param == "eig":
            _reseter(self, "eigenvalues")
            _reseter(self, "eigenvectors")
        if param == "matrix":
            _reseter(self, "matrix")
        if param == "all":
            self.reset("eig")
            self.reset("matrix")

    def solve_uniform(self, omega, kx, ky, nh):
        """Solve for eigenmodes and eigenvalues of a uniform layer.

        Parameters
        ----------
        omega : float
            Pulsation.
        kx : array_like
            Transverse wavenumber x component.
        ky : array_like
            Transverse wavenumber y component.
        nh : int
            Number of harmonics.

        Returns
        -------
        tuple
            `(q,psi)`, with eigenvalues `q` and eigenvectors `psi`.

        """
        epsilon = self.epsilon
        mu = self.mu

        if self.is_mu_anisotropic or self.is_epsilon_anisotropic:
            # TODO: anisotropic uniform layer
            raise NotImplementedError("Uniform layer material must be isotropic")

        Id = bk.eye(nh)
        if self.is_epsilon_anisotropic:
            _epsilon = block(
                [
                    [epsilon[0, 0] * Id, epsilon[0, 1] * Id],
                    [epsilon[1, 0] * Id, epsilon[1, 1] * Id],
                ]
            )
        else:
            _epsilon = epsilon

        if self.is_mu_anisotropic:
            _mu = block(
                [[mu[0, 0] * Id, mu[0, 1] * Id], [mu[1, 0] * Id, mu[1, 1] * Id]]
            )
        else:
            _mu = mu

        q = (
            bk.array(
                _epsilon * _mu * omega**2 - kx**2 - ky**2,
                dtype=bk.complex128,
            )
            ** 0.5
        )
        # q = bk.where(bk.imag(q) < 0.0, -q, q)
        self.eigenvalues = bk.hstack((q, q))
        self.eigenvectors = bk.eye(2 * len(kx), dtype=bk.complex128)
        return self.eigenvalues, self.eigenvectors

    def solve_eigenproblem(self, matrix):
        """Solve the eigenproblem for a patterned layer.

        Parameters
        ----------
        matrix : array_like
            The matrix which we search for eigenvalues and eigenvectors.

        Returns
        -------
        tuple
            `(q,psi)`, with eigenvalues `q` and eigenvectors `psi`.

        """
        if self.iscopy:
            self.eigenvalues, self.eigenvectors = (
                self.original.eigenvalues,
                self.original.eigenvectors,
            )

        else:
            # # TODO: implement custom autodiff rules for the evp with jax
            # if BACKEND == "jax":
            #     # from ._jax_eig_workaround import eig_jax
            #     eig_func = bk.linalg.eig
            # elif BACKEND == "torch":
            #     from . import torch
            #
            #     def eig_func(u):
            #         u = torch.tensor(u)#.to(_device)
            #         return torch.linalg.eig(u)
            #
            # else:
            #     eig_func = bk.linalg.eig
            eig_func = jit(bk.linalg.eig)

            w, v = eig_func(matrix)
            q = w**0.5
            q = bk.where(bk.imag(q) < 0.0, -q, q)
            self.eigenvalues, self.eigenvectors = q, v
        return self.eigenvalues, self.eigenvectors

    def copy(self, name=None):
        """Copy a layer.

        Returns
        -------
        nannos.Layer
            A copy of the layer.

        """
        cp = copy(self)
        cp.iscopy = True
        cp.original = self
        if name is None:
            cp.name += " (copy)"
        else:
            cp.name = name
        return cp

    @property
    def is_epsilon_anisotropic(self):
        return is_anisotropic(self.epsilon)

    @property
    def is_mu_anisotropic(self):
        return is_anisotropic(self.mu)

    @property
    def is_uniform(self):
        """Check if layer is uniform.

        Returns
        -------
        bool
            ``True`` if the layer is uniform, ``False`` if not.

        """
        if self.is_epsilon_anisotropic:
            e = self.epsilon.shape == (3, 3)
        else:
            e = self.epsilon.shape == ()
        if self.is_mu_anisotropic:
            m = self.mu.shape == (3, 3)
        else:
            m = self.mu.shape == ()

        return e and m

    def get_tangent_field(self, harmonics, normalize=False, type=None):
        type = type or self.tangent_field_type
        if self.is_uniform:
            return None
        else:
            if self.tangent_field is not None:
                return self.tangent_field
            else:
                epsilon = self.epsilon
                epsilon_zz = epsilon[2, 2] if self.is_epsilon_anisotropic else epsilon
                return get_tangent_field(
                    epsilon_zz,
                    harmonics,
                    normalize=normalize,
                    type=self.tangent_field_type,
                )

    def get_jones_field(self, t):
        if self.is_uniform:
            return None
        else:
            return get_jones_field(t)


def is_anisotropic(f):
    return f.shape[:2] == (3, 3)


def _get_layer(id, layers, layer_names):
    """Helper to get layer index and name.

    Parameters
    ----------
    id : int or str
        The index of the layer or its name.
    layers : list of Layers objects
        The layer list (stack).
    id : list of str
        Names of the layers.

    Returns
    -------
    layer : nannos.Layer
        The layer object.
    index : nannos.Layer
        The layer index in the stack.
    """
    if isinstance(id, str):
        if id in layer_names:
            for i, l in enumerate(layers):
                if l.name == id:
                    return l, i
        else:
            raise ValueError(f"Unknown layer name {id}")
    elif isinstance(id, int):
        return layers[id], id
    elif hasattr(id, "name"):
        return _get_layer(id.name, layers, layer_names)
    else:
        raise ValueError(
            f"Wrong id for layer: {id}. Please use an integrer specifying the layer index or a string for the layer name."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import numpy as npo

from .. import backend as bk
from ..utils import apply_filter, norm
from .fft import fourier_transform, inverse_fourier_transform


def _normalize(x, n, invalid=0, threshold=0.0):
    with npo.errstate(invalid="ignore"):
        f = x / (n)
    return bk.array(bk.where(n <= threshold, invalid * bk.ones_like(x), f))


def _normalize_vec(V, n, threshold=1e-6):
    Vx = _normalize(V[0], n, invalid=1, threshold=threshold)
    Vy = _normalize(V[1], n, invalid=0, threshold=threshold)
    return [Vx, Vy]


def _ft_filt(x, expo):
    with npo.errstate(divide="ignore"):
        f = 1 / (x**expo)
    return bk.array(bk.where(x == 0.0, 0.0 * x, f))


def _get_tangent_field_fft(grid, normalize=False, rfilt=4, expo=0.5):
    Nx, Ny = grid.shape

    grid = bk.hstack([grid, grid, grid])
    grid = bk.vstack([grid, grid, grid])

    xf = apply_filter(grid, rfilt)
    v = bk.gradient(grid)
    vf = bk.gradient(xf)
    norma = norm(v)
    N = [bk.array(v[i]) * bk.abs(vf[i]) for i in range(2)]
    N = [_normalize(N[i], norma) for i in range(2)]
    fx = bk.fft.fftfreq(3 * Nx)
    fy = bk.fft.fftfreq(3 * Ny)
    Fx, Fy = bk.meshgrid(fx, fy, indexing="ij")
    ghat = _ft_filt(Fx**2 + Fy**2, expo=expo) * Nx * Ny
    # ghat = apply_filter(Fx**2 + Fy**2, rfilt)
    Nhat = [fourier_transform(N[i]) for i in range(2)]
    Nstar = [(inverse_fourier_transform(Nhat[i] * ghat)) for i in range(2)]

    Nstar = [Nstar[i][Nx : 2 * Nx, Ny : 2 * Ny] for i in range(2)]
    t = [Nstar[1].real, -Nstar[0].real]
    # t = bk.array(t).real
    if normalize:
        norm_t = norm(t)
        t = _normalize_vec(t, norm_t)
    else:
        t = [t[i] / bk.max(norm(t)) for i in range(2)]
    return t


def _get_tangent_field_min(
    grid, harmonics, normalize=False, rfilt=4, opt_backend="autograd"
):

    if opt_backend == "jax":
        from jax import grad
        from jax import numpy as npg
        from jax.config import config

        # config.update("jax_platform_name", "cpu")
        config.update("jax_enable_x64", True)
        from jax import jit
    else:
        from autograd import grad
        from autograd import numpy as npg

        def jit(x):
            return x

    from scipy.optimize import minimize

    def norm(v):
        # avoid division by 0
        eps = npg.finfo(float).eps
        return npg.sqrt(eps + v[0] * npg.conj(v[0]) + v[1] * npg.conj(v[1]))

    def _normalize(x, n):
        with npo.errstate(invalid="ignore"):
            f = x / (n)
        return npg.array(npg.where(n == 0.0, 0.0 * x, f))

    Nx, Ny = grid.shape
    Nx_ds = min(2**6, Nx)
    Ny_ds = min(2**6, Ny)

    downsample_x = int(Nx / Nx_ds)
    downsample_y = int(Ny / Ny_ds)
    shape_small = (Nx_ds, Ny_ds)

    nh = len(harmonics[0])
    nh = min(nh, 51)
    harmonics = harmonics[:, :nh]

    xf = apply_filter(grid, rfilt=rfilt)
    dgrid_f = npg.array(npg.gradient(xf))

    normdgrid_f = norm(dgrid_f)
    # maxi = npg.max(normdgrid_f)
    # dgrid_f = npg.array([dgrid_f[i] /maxi for i in range(2)])
    dgrid_f = npg.array([_normalize(dgrid_f[i], normdgrid_f) for i in range(2)])

    def _set_idx(mat, idx, val):
        if opt_backend == "jax":
            mat = mat.at[tuple(idx)].set(val)
        else:
            # idx += [None]
            mat[tuple(idx)] = val
        return mat

    def _get_amps(amplitudes, shape):
        amplitudes = npg.array([amplitudes])
        if len(amplitudes.shape) == 1:
            amplitudes = npg.reshape(amplitudes, amplitudes.shape + (1,))
        f = npg.zeros(shape + (amplitudes.shape[0],) + (nh,), dtype=npg.complex128)
        # f[harmonics[0], harmonics[1],0] = npg.eye(nh)
        idx = [harmonics[0], harmonics[1], 0]
        # f[tuple(idx)] =  npg.eye(nh)
        f = _set_idx(f, idx, npg.eye(nh))
        s = npg.sum(amplitudes * f, axis=-1)
        ft = npg.fft.ifft2(s, axes=(0, 1)) * Nx * Ny
        return ft[:, :, 0]

    @jit
    def get_amps(amplitudes):
        return _get_amps(amplitudes, shape_small)

    def minifun(x):
        coef = x[:nh] + 1j * x[nh:]
        a = get_amps(coef)
        da = npg.array(npg.gradient(a))
        obj = npg.mean(npg.abs(da - dgrid_f[:, ::downsample_x, ::downsample_y]) ** 2)
        return obj

    @jit
    def minifun_der(x):
        d = grad(minifun)(x)
        return d

    x0 = npg.zeros(2 * nh)
    # x0 = npo.random.rand(2 * nh)
    res = minimize(
        minifun,
        x0,
        method="BFGS",
        jac=minifun_der,
        tol=5e-4,
        options={"disp": False, "maxiter": 50},
    )

    xopt = res.x
    coef = xopt[:nh] + 1j * xopt[nh:]

    a = _get_amps(coef, (Nx, Ny))
    n = npg.array(npg.gradient(a))
    t = [n[1], -n[0]]

    t = bk.array(t).real

    if normalize:
        norm_t = norm(t)
        t = _normalize_vec(t, norm_t)
    else:
        t = [t[i] / bk.max(norm(t)) for i in range(2)]

    return t


def get_tangent_field(grid, harmonics, normalize=False, rfilt=4, type="fft"):
    if type == "fft":
        return _get_tangent_field_fft(grid, normalize, rfilt=rfilt)
    elif type == "opt":
        return _get_tangent_field_min(grid, harmonics, normalize, rfilt=rfilt)
    else:
        raise ValueError(
            f"Wrong type of tangent field {type}. Please choose 'fft' or 'opt'"
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io


import nlopt
import numpy as npo
from scipy.optimize import minimize

from . import DEVICE
from . import backend as bk
from . import get_backend, grad
from .utils import apply_filter, tic, toc


def simp(x, eps_min, eps_max, p=1):
    return (eps_max - eps_min) * x**p + eps_min


def project(x, beta=1, thres=0.5):
    x = bk.array(x)

    def tanh(z):
        return bk.tanh(bk.array(z))

    return ((tanh(thres * beta)) + tanh(beta * (x - thres))) / (
        tanh(thres * beta) + (tanh((1 - thres) * beta))
    )


def multi_project(x, beta=1, Nthres=2):
    thresholds = bk.linspace(0, 1, Nthres + 1)[1:-1]
    out = 0
    for thres in thresholds:
        out += project(x, beta, thres)
    return out / len(thresholds)


def multi_simp(x, epsilons, p=1):
    epsilons = bk.array(epsilons) + 0j
    nthres = len(epsilons)
    npts = nthres
    if nthres == 2:
        return simp(x, epsilons[0], epsilons[1], p)
    else:
        pts = bk.linspace(0, 1, npts)

        def pol(coefs, x):
            return sum(c * x ** (n * p) for n, c in enumerate(coefs))

        def mat(pts):
            return bk.array([[_x ** (n * p) + 0j for n in range(npts)] for _x in pts])

        M = mat(pts)
        coefs = bk.linalg.inv(M) @ epsilons
        return pol(coefs, x)


class StopFunError(Exception):
    """Raised when stop value reached"""

    pass


class StopFun:
    def __init__(self, fun, stopval=None):
        self.fun_in = fun
        self.stopval = stopval

    def fun(self, x, *args):
        self.f = self.fun_in(x, *args)
        if self.stopval is not None:
            if self.f < self.stopval:
                raise StopFunError("Stop value reached.")
        self.x = x
        return self.f


class TopologyOptimizer:
    def __init__(
        self,
        fun,
        x0,
        method="scipy",
        threshold=(0, 8),
        maxiter=10,
        stopval=None,
        args=None,
        callback=None,
        verbose=False,
        options={},
    ):
        self.fun = fun
        self.x0 = x0
        self.nvar = len(x0)
        self.method = method
        self.threshold = threshold
        self.maxiter = maxiter
        self.stopval = stopval
        self.args = args
        self.callback = callback
        self.verbose = verbose
        self.options = options
        self.grad_fun = grad(fun)

        self.current_iteration = 0

    def print(self, s):
        if self.verbose:
            return print(s)

    def minimize(self):

        self.print("#################################################")
        self.print(f"Topology optimization with {self.nvar} variables")
        self.print("#################################################")
        self.print("")
        x0 = (
            self.x0.cpu()
            if (get_backend() == "torch" and DEVICE == "cuda")
            else self.x0
        )
        x0 = npo.array(x0)
        for iopt in range(*self.threshold):
            self.print("-----------------------")
            self.print(f"  global iteration {iopt}")
            self.print("-----------------------")

            proj_level = 2**iopt
            args = list(self.args)
            args[0] = proj_level
            args = tuple(args)
            if self.method == "scipy":

                def fun_scipy(x, *args):
                    self.print(f">>> iteration {self.current_iteration}")
                    x = bk.array(x, dtype=bk.float64)
                    y = self.fun(x, *args)
                    self.print(f"current value = {y}")
                    if self.callback is not None:
                        self.callback(x, y, *args)
                    self.current_iteration += 1
                    return y

                bounds = [(0, 1) for _ in range(self.nvar)]
                options = {"maxiter": self.maxiter}
                if self.options is not None:
                    options.update(self.options)

                stop = StopFun(fun=fun_scipy, stopval=self.stopval)
                try:
                    self.opt = minimize(
                        stop.fun,
                        x0,
                        args=args,
                        method="L-BFGS-B",
                        jac=self.grad_fun,
                        bounds=bounds,
                        options=options,
                    )
                    xopt = self.opt.x
                    fopt = self.opt.fun
                except StopFunError as e:
                    print(e)
                    xopt = stop.x
                    fopt = stop.f

            else:

                def fun_nlopt(x, gradn):
                    self.print(f">>> iteration {self.current_iteration}")
                    x = bk.array(x, dtype=bk.float64)
                    y = self.fun(x, *args)
                    self.print(f"current value = {y}")
                    if self.callback is not None:
                        self.callback(x, y, *args)
                    if gradn.size > 0:
                        dy = self.grad_fun(x, *args)
                        dy = (
                            dy.cpu()
                            if (get_backend() == "torch" and DEVICE == "cuda")
                            else dy
                        )
                        gradn[:] = npo.array(dy, dtype=npo.float64)

                    self.current_iteration += 1
                    return npo.float(y)

                lb = npo.zeros(self.nvar, dtype=npo.float64)
                ub = npo.ones(self.nvar, dtype=npo.float64)

                self.opt = nlopt.opt(nlopt.LD_MMA, self.nvar)
                self.opt.set_lower_bounds(lb)
                self.opt.set_upper_bounds(ub)
                if "ftol_rel" in self.options:
                    self.opt.set_ftol_rel(self.options["ftol_rel"])
                if "xtol_rel" in self.options:
                    self.opt.set_xtol_rel(self.options["xtol_rel"])
                if "ftol_abs" in self.options:
                    self.opt.set_ftol_abs(self.options["ftol_abs"])
                if "xtol_abs" in self.options:
                    self.opt.set_xtol_abs(self.options["xtol_abs"])
                if self.stopval is not None:
                    self.opt.set_stopval(self.stopval)
                if self.maxiter is not None:
                    self.opt.set_maxeval(self.maxiter)
                for k, v in self.options.items():
                    self.opt.set_param(k, v)
                self.opt.set_min_objective(fun_nlopt)
                xopt = self.opt.optimize(x0)
                fopt = self.opt.last_optimum_value()
            x0 = xopt
        return xopt, fopt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of nannos
# License: GPLv3
# See the documentation at nannos.gitlab.io

from . import backend as bk
from .constants import pi

__all__ = ["PlaneWave"]

_deg2rad = pi / 180


class PlaneWave:
    r"""A plane wave object

    Parameters
    ----------
    wavelength : float
        Wavelength (the default is 1).
    angles : tuple
        Incidence angles :math:`(\theta,\phi,\psi)` in degrees (the default is (0, 0, 0)).
        :math:`\\theta`: polar angle,
        :math:`\phi`: azimuthal angle,
        :math:`\psi`: polarization angle.

    """

    def __init__(self, wavelength=1, angles=(0, 0, 0)):
        self.wavelength = bk.array(wavelength)
        self.angles_deg = bk.array(angles, dtype=bk.float64)
        self.angles = self.angles_deg * _deg2rad
        self.theta = bk.array(self.angles[0], dtype=bk.float64)
        self.phi = bk.array(self.angles[1], dtype=bk.float64)
        self.psi = bk.array(self.angles[2], dtype=bk.float64)
        self.frequency_scaled = bk.array(1 / self.wavelength)

        k0 = 2 * pi * self.frequency_scaled

        self.wavenumber = k0
        self.wavevector = k0 * bk.array(
            [
                bk.sin(self.theta) * bk.cos(self.phi),
                bk.sin(self.theta) * bk.sin(self.phi),
                bk.cos(self.theta),
            ]
        )

        cpsi = bk.cos(self.psi)  # p
        spsi = bk.sin(self.psi)  # s

        # cx = cpsi * bk.cos(self.theta) * bk.cos(self.phi) - spsi * bk.sin(self.phi)
        # cy = cpsi * bk.cos(self.theta) * bk.sin(self.phi) + spsi * bk.cos(self.phi)
        # cz = cpsi * bk.sin(self.theta)
        cx = cpsi * bk.cos(self.theta) * bk.cos(self.phi) - spsi * bk.sin(self.phi)
        cy = cpsi * bk.cos(self.theta) * bk.sin(self.phi) + spsi * bk.cos(self.phi)
        cz = -cpsi * bk.sin(self.theta)

        self.amplitude = bk.array([cx, cy, cz])

        kt = self.wavevector
        omega = k0
        K = bk.array([[kt[1] ** 2, -kt[0] * kt[1]], [-kt[0] * kt[1], kt[0] ** 2]])
        Q = omega**2 * bk.array(bk.eye(2)) - K
        q = (omega**2 - kt[0] ** 2 - kt[1] ** 2) ** 0.5
        C = bk.linalg.inv(Q) * (omega * q)
        et = bk.array([-self.amplitude[1], self.amplitude[0]])
        self.a0 = C @ et
        # self.a0 = -self.amplitude[1], self.amplitude[0]
        # self.a0 = self.amplitude[0], self.amplitude[1]
